import logging
import os
from pathlib import Path
from typing import Set

import click
from exasol_integration_test_docker_environment.lib.base import luigi_log_config
from exasol_integration_test_docker_environment.lib.config import build_config

from exasol_script_languages_container_ci.lib.branch_config import BranchConfig
from exasol_script_languages_container_ci.lib.common import get_config
from exasol_script_languages_container_ci.lib.ci_build import CIBuild
from exasol_script_languages_container_ci.lib.ci_push import CIPush
from exasol_script_languages_container_ci.lib.ci_security_scan import CISecurityScan
from exasol_script_languages_container_ci.lib.ci_test import CIExecuteTest
from exasol_script_languages_container_ci.lib.git_access import GitAccess


def get_all_affected_files(git_access: GitAccess, base_branch: str) -> Set[str]:
    base_last_commit_sha = git_access.get_head_commit_sha_of_branch(base_branch)
    changed_files = set()
    for commit in git_access.get_last_commits():
        if commit == base_last_commit_sha:
            break
        changed_files.update(git_access.get_files_of_commit(commit))
    return changed_files


def check_if_need_to_build(branch_name: str, config_file: str, flavor: str, git_access: GitAccess):
    if BranchConfig.build_always(branch_name):
        return True
    if "[rebuild]" in git_access.get_last_commit_message():
        return True
    with get_config(config_file) as config:
        affected_files = list(get_all_affected_files(git_access, config["base_branch"]))
        logging.debug(f"check_if_need_to_build: Found files of last commits: {affected_files}")
        for ignore_path in config["build_ignore"]["ignored_paths"]:
            affected_files = list(filter(lambda file: not file.startswith(ignore_path), affected_files))

    if len(affected_files) > 0:
        # Now filter out also other flavor folders
        this_flavor_path = f"flavors/{flavor}"
        affected_files = list(filter(lambda file: not file.startswith("flavors") or file.startswith(this_flavor_path),
                                     affected_files))
    logging.debug(f"check_if_need_to_build: filtered files: {affected_files}")
    return len(affected_files) > 0


def ci(flavor: str,
       branch_name: str,
       docker_user: str,
       docker_password: str,
       docker_build_repository: str,
       docker_release_repository: str,
       commit_sha: str,
       config_file: str,
       git_access: GitAccess,
       ci_build: CIBuild = CIBuild(),
       ci_execute_tests: CIExecuteTest = CIExecuteTest(),
       ci_push: CIPush = CIPush(),
       ci_security_scan: CISecurityScan = CISecurityScan()):
    """
    Run CI build:
    1. Build image
    2. Run db tests
    3. Run security scan
    4. Push to docker repositories
    """
    logging.info(f"Running CI build for parameters: {locals()}")

    flavor_path = (f"flavors/{flavor}",)
    test_container_folder = "test_container"
    rebuild = BranchConfig.rebuild(branch_name)
    needs_to_build = check_if_need_to_build(branch_name, config_file, flavor, git_access)
    if needs_to_build:
        log_path = Path(build_config.DEFAULT_OUTPUT_DIRECTORY) / "jobs" / "logs" / "main.log"
        os.environ[luigi_log_config.LOG_ENV_VARIABLE_NAME] = f"{log_path.absolute()}"

        ci_build.build(flavor_path=flavor_path,
                       rebuild=rebuild,
                       build_docker_repository=docker_build_repository,
                       commit_sha=commit_sha,
                       docker_user=docker_user,
                       docker_password=docker_password,
                       test_container_folder=test_container_folder)
        ci_execute_tests.execute_tests(flavor_path=flavor_path,
                                       docker_user=docker_user,
                                       docker_password=docker_password,
                                       test_container_folder=test_container_folder)
        ci_security_scan.run_security_scan(flavor_path=flavor_path)
        ci_push.push(flavor_path=flavor_path,
                     target_docker_repository=docker_build_repository,
                     target_docker_tag_prefix=commit_sha,
                     docker_user=docker_user,
                     docker_password=docker_password)
        ci_push.push(flavor_path=flavor_path,
                     target_docker_repository=docker_build_repository,
                     target_docker_tag_prefix="",
                     docker_user=docker_user,
                     docker_password=docker_password)
        if BranchConfig.push_to_docker_release_repo(branch_name):
            ci_push.push(flavor_path=flavor_path,
                         target_docker_repository=docker_release_repository,
                         target_docker_tag_prefix="",
                         docker_user=docker_user,
                         docker_password=docker_password)
    else:
        logging.warning(f"Skipping build...")



import logging
from typing import Tuple

import click
from exasol_script_languages_container_tool.lib.api.run_db_tests import run_db_test

from exasol_script_languages_container_ci.lib.common import print_docker_images

from exasol_script_languages_container_tool.lib.tasks.test.test_container import AllTestsResult


class CIExecuteTest:

    def execute_tests(self,
                      flavor_path: Tuple[str, ...],
                      docker_user: str,
                      docker_password: str,
                      test_container_folder: str) -> Tuple[AllTestsResult, AllTestsResult]:
        """
        Run db tests
        """
        logging.info(f"Running command 'run_db_test' for flavor-path {flavor_path}")
        run_db_test_flavor_result = \
            run_db_test(flavor_path=flavor_path,
                        workers=7,
                        source_docker_username=docker_user,
                        source_docker_password=docker_password,
                        test_container_folder=test_container_folder)
        logging.info(f"Running command 'run_db_test' for linker_namespace_sanity for flavor-path {flavor_path}")
        run_db_test_linkernamespace = \
            run_db_test(flavor_path=flavor_path, workers=7,
                        test_folder=("test/linker_namespace_sanity",),
                        release_goal=("base_test_build_run",),
                        source_docker_username=docker_user,
                        source_docker_password=docker_password,
                        test_container_folder=test_container_folder)
        print_docker_images(logging.info)
        return run_db_test_flavor_result, run_db_test_linkernamespace



import logging
from typing import Tuple

import click
from exasol_script_languages_container_tool.lib.api.push import push

from exasol_script_languages_container_ci.lib.common import print_docker_images


class CIPush:
    def push(self,
             flavor_path: Tuple[str, ...],
             target_docker_repository: str,
             target_docker_tag_prefix: str,
             docker_user: str,
             docker_password: str):
        """
        Push the docker image to Dockerhub
        """

        logging.info(f"Running command 'push' with parameters: {locals()}")
        push(flavor_path=flavor_path,
             push_all=True,
             force_push=True,
             workers=7,
             target_docker_repository_name=target_docker_repository,
             target_docker_tag_prefix=target_docker_tag_prefix,
             target_docker_username=docker_user,
             target_docker_password=docker_password)
        print_docker_images(logging.info)



import json
from contextlib import contextmanager
from pathlib import Path
from typing import Callable
from inspect import cleandoc

import docker


def _get_docker_images():
    docker_client = docker.from_env()
    exa_images = []
    try:
        exa_images = [str(img) for img in docker_client.images.list() if "exasol" in str(img)]
    finally:
        docker_client.close()
    return exa_images


def print_docker_images(writer: Callable[[str], None]):
    """
    Prints all docker images whith "exa" in it's name to stdout.
    :return: None
    """

    writer(cleandoc("""
        {seperator}
        Printing docker images
        {seperator}
        {images}""").format(
        seperator=20 * "=", images="\n".join(_get_docker_images())
    ))


def print_file(filename: Path, writer: Callable[[str], None]):
    """
    Opens file readonly, reads it content and prints to writer.
    """
    with open(filename, "r") as f:
        writer(f.read())


@contextmanager
def get_config(config_file: str):
    """
    Opens config file and returns parsed JSON object.
    """
    with open(config_file, "r") as f:
        yield json.load(f)



import logging
from typing import Tuple

from exasol_script_languages_container_tool.lib.api import export
from exasol_script_languages_container_tool.lib.tasks.export.export_containers import ExportContainerResult

from exasol_script_languages_container_ci.lib.common import print_docker_images


class CIExport:
    def export(self,
               flavor_path: Tuple[str, ...],
               export_path: str) -> ExportContainerResult:
        """
        Export the flavor as tar.gz file
        """

        logging.info(f"Running command 'push' with parameters: {locals()}")
        export_result = export(flavor_path=flavor_path,
                               export_path=export_path,
                               workers=7)
        print_docker_images(logging.info)
        return export_result



import logging
from pathlib import Path

from github import Github, GithubException


class GithubReleaseAssetUploader(object):
    """
    Implements upload to a Github Release.
    See https://docs.github.com/en/rest/releases/assets#upload-a-release-asset for details.
    The access token needs to be stored in the environment variable GITHUB_TOKEN
    """
    def __init__(self, token):
        self._token = token

    def upload(self, archive_path: str, label: str, repo_id: str, release_id: int, content_type: str):
        gh = Github(self._token)
        gh_repo = gh.get_repo(repo_id)
        release = gh_repo.get_release(release_id)
        # Check GH limitation
        # https://docs.github.com/en/repositories/releasing-projects-on-github/about-releases#storage-and-bandwidth-quotas
        if Path(archive_path).stat().st_size >= 2 * (2 ** 30):
            logging.error("File larger than 2GB. Skipping it...")
        else:
            try:
                release.upload_asset(path=archive_path, label=label, content_type=content_type)
            except GithubException as ex:
                logging.error(f"Upload of asset {archive_path} to release {release_id} failed: {ex}")
                raise ex



import logging
from pathlib import Path
from typing import Tuple

from exasol_script_languages_container_tool.lib.api import security_scan
from exasol_script_languages_container_tool.lib.tasks.security_scan.security_scan import AllScanResult

import exasol_script_languages_container_ci.lib.common


class CISecurityScan:

    def run_security_scan(self, flavor_path: Tuple[str, ...]) -> AllScanResult:
        """
        Run security scan and print result
        """

        logging.info(f"Running command 'security_scan' with parameters {locals()}")
        security_scan_result = security_scan(flavor_path=flavor_path, workers=7)
        exasol_script_languages_container_ci.lib.common.print_docker_images(logging.info)
        logging.info("============= SECURITY REPORT ===========")
        # Important: Call print_file over the global module name, otherwise the patch in the unit-test does not work!
        exasol_script_languages_container_ci.lib.common.print_file(
            Path() / ".build_output" / "security_scan" / "security_report", logging.info)
        return security_scan_result



import logging

import click

from exasol_script_languages_container_ci.cli.cli import cli
from exasol_script_languages_container_ci.lib.ci import ci
from exasol_script_languages_container_ci.lib.git_access import GitAccess


@cli.command()
@click.option('--flavor', required=True, type=str,
              help="Flavor name.")
@click.option('--branch-name', required=True, type=str,
              help="Branch name.")
@click.option('--docker-user', required=True, type=str,
              help="Docker user name")
@click.option('--docker-password', required=True, type=str,
              help="Docker password")
@click.option('--docker-build-repository', required=True, type=str,
              help="Docker build repository")
@click.option('--docker-release-repository', required=True, type=str,
              help="Docker release repository")
@click.option('--commit-sha', required=True, type=str,
              help="Commit SHA")
@click.option('--config-file', required=True, type=click.Path(exists=True, file_okay=True, dir_okay=False),
              help="The build config file (project specific)")
@click.pass_context
def run_ci(ctx: click.Context,
           flavor: str,
           branch_name: str,
           docker_user: str,
           docker_password: str,
           docker_build_repository: str,
           docker_release_repository: str,
           commit_sha: str,
           config_file: str):
    logging.basicConfig(level=logging.INFO)
    ci(flavor,
       branch_name,
       docker_user,
       docker_password,
       docker_build_repository,
       docker_release_repository,
       commit_sha,
       config_file,
       GitAccess())



import logging
from typing import Tuple, Dict, Optional

from exasol_integration_test_docker_environment.lib.api.build_test_container import build_test_container
from exasol_integration_test_docker_environment.lib.docker.images.image_info import ImageInfo
from exasol_script_languages_container_tool.lib.api import build
from exasol_script_languages_container_tool.lib.tasks.test.test_container_content import build_test_container_content

from exasol_script_languages_container_ci.lib.common import print_docker_images


class CIBuild:
    def build(self,
              flavor_path: Tuple[str, ...],
              rebuild: bool,
              build_docker_repository: Optional[str],
              commit_sha: str,
              docker_user: str,
              docker_password: str,
              test_container_folder: str) -> Tuple[Dict[str, Dict[str, ImageInfo]], ImageInfo]:
        """
        Build the script-language container for given flavor. And also build the test container
        """

        logging.info(f"Running command 'build' with parameters: {locals()}")
        if build_docker_repository is None:
            slc_image_infos = build(flavor_path=flavor_path, force_rebuild=rebuild,
                                    source_docker_tag_prefix=commit_sha,
                                    source_docker_username=docker_user,
                                    source_docker_password=docker_password,
                                    shortcut_build=False,
                                    workers=7)
        else:
            slc_image_infos = build(flavor_path=flavor_path, force_rebuild=rebuild,
                                    source_docker_repository_name=build_docker_repository,
                                    source_docker_tag_prefix=commit_sha,
                                    source_docker_username=docker_user,
                                    source_docker_password=docker_password,
                                    shortcut_build=False,
                                    workers=7)
        logging.info(f"Running command 'build_test_container' with parameters: {locals()}")
        content = build_test_container_content(test_container_folder)
        test_container_image_infos = build_test_container(
            force_rebuild=rebuild,
            workers=7,
            test_container_content=content,
            source_docker_repository_name=build_docker_repository,
            source_docker_tag_prefix=commit_sha
        )
        print_docker_images(logging.info)
        return slc_image_infos, test_container_image_infos



import glob
import logging
from pathlib import Path

from exasol_script_languages_container_ci.lib.github_release_asset_uploader import GithubReleaseAssetUploader


class AssetUploader:

    def __init__(self, release_asset_uploader: GithubReleaseAssetUploader):
        self._release_asset_uploader = release_asset_uploader

    def upload_assets(self,
                      repo_id: str, release_id: int, content_type: str,
                      artifact_path: str, file_suffix: str, label_prefix: str):
        release_artifacts = glob.glob(f'{artifact_path}/*{file_suffix}')
        for release_artifact in release_artifacts:
            artifact_file_name = Path(release_artifact).name
            if artifact_file_name.endswith(file_suffix):
                artifact_file_name = artifact_file_name[:-len(file_suffix)]
            else:
                logging.error(f"Artifact file: {artifact_file_name} does not end with {file_suffix}. "
                              f"Using {artifact_file_name} as label.")
            self._release_asset_uploader.upload(archive_path=release_artifact,
                                                label=f"{label_prefix} {artifact_file_name}",
                                                repo_id=repo_id, release_id=release_id, content_type=content_type)



import logging
import os

import click

from exasol_script_languages_container_ci.cli.cli import cli
from exasol_script_languages_container_ci.lib.asset_uploader import AssetUploader
from exasol_script_languages_container_ci.lib.github_release_asset_uploader import GithubReleaseAssetUploader
from exasol_script_languages_container_ci.lib.release import release
from exasol_script_languages_container_ci.lib.release_uploader import ReleaseUploader


@cli.command()
@click.option('--flavor', required=True, type=str,
              help="Flavor name.")
@click.option('--docker-user', required=True, type=str,
              help="Docker user name")
@click.option('--docker-password', required=True, type=str,
              help="Docker password")
@click.option('--docker-release-repository', required=True, type=str,
              help="Docker release repository")
@click.option('--config-file', required=True, type=click.Path(exists=True, file_okay=True, dir_okay=False),
              help="The build config file (project specific)")
@click.option('--source-repo-url', required=True, type=str,
              help="The url of the repository. Usually set by AWS under env variable CODEBUILD_SOURCE_REPO_URL.")
@click.option('--release-id', required=True, type=int,
              help="The id of the release.")
@click.option('--dry-run/--no-dry-run', default=False,
              help="If true, runs release without pushing the container to the docker release repository."
                   "If false, also pushes the container to the docker release repository.")
@click.pass_context
def run_release(ctx: click.Context,
                flavor: str,
                docker_user: str,
                docker_password: str,
                docker_release_repository: str,
                config_file: str,
                source_repo_url: str,
                release_id: int,
                dry_run: bool):
    logging.basicConfig(level=logging.INFO)
    github_release_asset_uploader = GithubReleaseAssetUploader(os.getenv("GITHUB_TOKEN"))
    asset_uploader = AssetUploader(release_asset_uploader=github_release_asset_uploader)
    release_uploader = ReleaseUploader(asset_uploader=asset_uploader)
    release(flavor=flavor,
            docker_user=docker_user,
            docker_password=docker_password,
            docker_release_repository=docker_release_repository,
            config_file=config_file,
            source_repo_url=source_repo_url,
            release_id=release_id,
            release_uploader=release_uploader,
            is_dry_run=dry_run)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['exasol_script_languages_container_ci',
 'exasol_script_languages_container_ci.cli',
 'exasol_script_languages_container_ci.cli.commands',
 'exasol_script_languages_container_ci.lib']

package_data = \
{'': ['*']}

install_requires = \
['GitPython>=3.1.0',
 'PyGithub>=1.55.0,<2.0.0',
 'click>=8.0.3,<9.0.0',
 'exasol-integration-test-docker-environment>=1.4.0,<2.0.0',
 'exasol-script-languages-container-tool==0.16.0',
 'setuptools>=67.6.0,<68.0.0']

setup_kwargs = {
    'name': 'exasol-script-languages-container-ci',
    'version': '1.0.0',
    'description': 'Implements CI builds for script-language-container.',
    'long_description': 'None',
    'author': 'Thomas Uebensee',
    'author_email': 'ext.thomas.uebensee@exasol.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.0,<4.0',
}


setup(**setup_kwargs)



import logging
import os
from pathlib import Path

from exasol_integration_test_docker_environment.lib.base import luigi_log_config
from exasol_integration_test_docker_environment.lib.config import build_config

from exasol_script_languages_container_ci.lib.ci_build import CIBuild
from exasol_script_languages_container_ci.lib.ci_push import CIPush
from exasol_script_languages_container_ci.lib.ci_security_scan import CISecurityScan
from exasol_script_languages_container_ci.lib.ci_test import CIExecuteTest
from exasol_script_languages_container_ci.lib.release_uploader import ReleaseUploader


def release(flavor: str,
            docker_user: str,
            docker_password: str,
            docker_release_repository: str,
            config_file: str,
            source_repo_url: str,
            release_id: int,
            is_dry_run: bool,
            release_uploader: ReleaseUploader,
            ci_build: CIBuild = CIBuild(),
            ci_execute_tests: CIExecuteTest = CIExecuteTest(),
            ci_push: CIPush = CIPush(),
            ci_security_scan: CISecurityScan = CISecurityScan()):
    """
    Run Release build:
    1. Build image
    2. Run basic tests
    3. Push to docker release repository
    4. Upload to GH release url
    """
    logging.info(f"Running Release build for parameters: {locals()}")

    flavor_path = (f"flavors/{flavor}",)
    test_container_folder = "test_container"
    log_path = Path(build_config.DEFAULT_OUTPUT_DIRECTORY) / "jobs" / "logs" / "main.log"
    os.environ[luigi_log_config.LOG_ENV_VARIABLE_NAME] = f"{log_path.absolute()}"

    ci_build.build(flavor_path=flavor_path, rebuild=True,
                   build_docker_repository=None,
                   commit_sha="",
                   docker_user=None,
                   docker_password=None,
                   test_container_folder=test_container_folder)
    ci_execute_tests.execute_tests(flavor_path=flavor_path,
                                   docker_user=docker_user,
                                   docker_password=docker_password,
                                   test_container_folder=test_container_folder)
    ci_security_scan.run_security_scan(flavor_path=flavor_path)
    if not is_dry_run:
        ci_push.push(flavor_path=flavor_path,
                     target_docker_repository=docker_release_repository,
                     target_docker_tag_prefix="",
                     docker_user=docker_user,
                     docker_password=docker_password)
    else:
        logging.info("Skipping push to docker release repository due to dry-run.")
    release_uploader.release_upload(flavor_path=flavor_path,
                                    source_repo_url=source_repo_url,
                                    release_id=release_id)
